from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import math
from functools import partial

try:
    import accimage
except ImportError:
    accimage = None
import numpy as np


class RFTM:
    def __init__(self, device=None):
        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device

        self.model = generate_model(device)
        checkpoint = torch.load(
            "./weights/rftm_resnet101.pth", map_location=self.device
        )
        self.model.load_state_dict(checkpoint["state_dict"])
        self.model.eval()

        self.classifier = Learner().to(self.device)
        checkpoint = torch.load(
            "./weights/rftm_classifier.pth", map_location=self.device
        )
        self.classifier.load_state_dict(checkpoint["net"])

    def predict(self, inputs):
        inputs = inputs.to(self.device)
        _, feature = self.model(inputs)
        feature = F.normalize(feature, p=2, dim=1).to(self.device)
        out = self.classifier(feature)
        return out


def generate_model(device):
    model = resnet101(
        num_classes=400,
        shortcut_type="B",
        cardinality=32,
        sample_size=112,
        sample_duration=16,
        input_channels=3,
        output_layers=[],
    ).to(device)
    model = nn.DataParallel(model)
    return model


class Learner(nn.Module):
    def __init__(self, input_dim=2048, drop_p=0.0):
        super(Learner, self).__init__()
        self.classifier = nn.Sequential(
            nn.Linear(input_dim, 512),
            nn.ReLU(),
            nn.Dropout(0.6),
            nn.Linear(512, 32),
            nn.ReLU(),
            nn.Dropout(0.6),
            nn.Linear(32, 1),
            nn.Sigmoid(),
        )
        self.drop_p = 0.6
        self.weight_init()
        self.vars = nn.ParameterList()

        for i, param in enumerate(self.classifier.parameters()):
            self.vars.append(param)

    def weight_init(self):
        for layer in self.classifier:
            if type(layer) == nn.Linear:
                nn.init.xavier_normal_(layer.weight)

    def forward(self, x, vars=None):
        if vars is None:
            vars = self.vars
        #         print("x.shape : ", x.shape)
        #         print("vars[0].shape : ", vars[0].shape)
        #         print("vars[1].shape : ", vars[1].shape)
        x = F.linear(x, vars[0], vars[1])
        x = F.relu(x)
        x = F.dropout(x, self.drop_p, training=self.training)
        x = F.linear(x, vars[2], vars[3])
        x = F.dropout(x, self.drop_p, training=self.training)
        x = F.linear(x, vars[4], vars[5])
        return torch.sigmoid(x)

    def parameters(self):
        """
        override this function since initial parameters will return with a generator.
        :return:
        """
        return self.vars


__all__ = ["ResNeXt", "resnet50", "resnet101"]


def conv3x3x3(in_planes, out_planes, stride=1):
    # 3x3x3 convolution with padding
    return nn.Conv3d(
        in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False
    )


def downsample_basic_block(x, planes, stride):
    out = F.avg_pool3d(x, kernel_size=1, stride=stride)
    zero_pads = torch.Tensor(
        out.size(0), planes - out.size(1), out.size(2), out.size(3), out.size(4)
    ).zero_()
    if isinstance(out.data, torch.cuda.FloatTensor):
        zero_pads = zero_pads.cuda()

    out = Variable(torch.cat([out.data, zero_pads], dim=1))

    return out


class ResNeXtBottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, cardinality, stride=1, downsample=None):
        super(ResNeXtBottleneck, self).__init__()
        mid_planes = cardinality * int(planes / 32)
        self.conv1 = nn.Conv3d(inplanes, mid_planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm3d(mid_planes)
        self.conv2 = nn.Conv3d(
            mid_planes,
            mid_planes,
            kernel_size=3,
            stride=stride,
            padding=1,
            groups=cardinality,
            bias=False,
        )
        self.bn2 = nn.BatchNorm3d(mid_planes)
        self.conv3 = nn.Conv3d(
            mid_planes, planes * self.expansion, kernel_size=1, bias=False
        )
        self.bn3 = nn.BatchNorm3d(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNeXt(nn.Module):
    def __init__(
        self,
        block,
        layers,
        sample_size,
        sample_duration,
        shortcut_type="B",
        cardinality=32,
        num_classes=400,
        input_channels=3,
        output_layers=[],
    ):
        self.inplanes = 64
        super(ResNeXt, self).__init__()
        self.conv1 = nn.Conv3d(
            input_channels,
            64,
            kernel_size=7,
            stride=(1, 2, 2),
            padding=(3, 3, 3),
            bias=False,
        )

        self.bn1 = nn.BatchNorm3d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool3d(kernel_size=(3, 3, 3), stride=2, padding=1)
        self.layer1 = self._make_layer(
            block, 128, layers[0], shortcut_type, cardinality
        )
        self.layer2 = self._make_layer(
            block, 256, layers[1], shortcut_type, cardinality, stride=2
        )
        self.layer3 = self._make_layer(
            block, 512, layers[2], shortcut_type, cardinality, stride=2
        )
        self.layer4 = self._make_layer(
            block, 1024, layers[3], shortcut_type, cardinality, stride=2
        )
        last_duration = int(math.ceil(sample_duration / 16))
        last_size = int(math.ceil(sample_size / 32))
        self.avgpool = nn.AdaptiveAvgPool3d(1)
        self.fc = nn.Linear(cardinality * 32 * block.expansion, num_classes)

        # layer to output on forward pass
        self.output_layers = output_layers

        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                m.weight = nn.init.kaiming_normal_(m.weight, mode="fan_out")
            elif isinstance(m, nn.BatchNorm3d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, shortcut_type, cardinality, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            if shortcut_type == "A":
                downsample = partial(
                    downsample_basic_block,
                    planes=planes * block.expansion,
                    stride=stride,
                )
            else:
                downsample = nn.Sequential(
                    nn.Conv3d(
                        self.inplanes,
                        planes * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False,
                    ),
                    nn.BatchNorm3d(planes * block.expansion),
                )

        layers = []
        layers.append(block(self.inplanes, planes, cardinality, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, cardinality))

        return nn.Sequential(*layers)

    def forward(self, x):
        # pdb.set_trace()
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x1 = self.layer1(x)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)

        x5 = self.avgpool(x4)

        x6 = x5.view(x5.size(0), -1)
        x7 = self.fc(x6)

        if len(self.output_layers) == 0:
            return x7, x6
        else:
            out = []
            out.append(x7)
            for i in self.output_layers:
                if i == "avgpool":
                    out.append(x6)
                if i == "layer4":
                    out.append(x4)
                if i == "layer3":
                    out.append(x3)

        return out

    def freeze_batch_norm(self):
        for name, m in self.named_modules():
            if isinstance(m, nn.BatchNorm2d) or isinstance(
                m, nn.BatchNorm3d
            ):  # PHIL: i Think we can write just  "if isinstance(m, nn._BatchNorm)
                m.eval()  # use mean/variance from the training
                m.weight.requires_grad = False
                m.bias.requires_grad = False


def get_fine_tuning_parameters(model, ft_begin_index):
    if ft_begin_index == 0:
        return model.parameters()

    ft_module_names = []
    for i in range(ft_begin_index, 5):
        ft_module_names.append("layer{}".format(i))
    ft_module_names.append("fc")

    print("Layers to finetune : ", ft_module_names)

    parameters = []
    for k, v in model.named_parameters():
        for ft_module in ft_module_names:
            if ft_module in k:
                parameters.append({"params": v})
                break
        else:
            parameters.append({"params": v, "lr": 0.0})

    return parameters


def resnet50(**kwargs):
    """Constructs a ResNet-50 model."""
    model = ResNeXt(ResNeXtBottleneck, [3, 4, 6, 3], **kwargs)
    return model


def resnet101(**kwargs):
    """Constructs a ResNet-101 model."""
    model = ResNeXt(ResNeXtBottleneck, [3, 4, 23, 3], **kwargs)
    return model


def resnet152(**kwargs):
    """Constructs a ResNet-101 model."""
    model = ResNeXt(ResNeXtBottleneck, [3, 8, 36, 3], **kwargs)
    return model


class ToTensor(object):

    """Convert a ``PIL.Image`` or ``numpy.ndarray`` to tensor.
    Converts a PIL.Image or numpy.ndarray (H x W x C) in the range
    [0, 255] to a torch.FloatTensor of shape (C x H x W) in the range [0.0, 1.0].
    """

    def __init__(self, norm_value=255):
        self.norm_value = norm_value

    def __call__(self, pic):
        """
        Args:
            pic (PIL.Image or numpy.ndarray): Image to be converted to tensor.
        Returns:
            Tensor: Converted image.
        """
        if isinstance(pic, np.ndarray):
            # handle numpy array
            img = torch.from_numpy(pic.transpose((2, 0, 1)))
            # backward compatibility
            return img.float().div(self.norm_value)

        if accimage is not None and isinstance(pic, accimage.Image):
            nppic = np.zeros([pic.channels, pic.height, pic.width], dtype=np.float32)
            pic.copyto(nppic)
            return torch.from_numpy(nppic)

        # handle PIL Image
        if pic.mode == "I":
            img = torch.from_numpy(np.array(pic, np.int32, copy=False))
        elif pic.mode == "I;16":
            img = torch.from_numpy(np.array(pic, np.int16, copy=False))
        else:
            img = torch.ByteTensor(torch.ByteStorage.from_buffer(pic.tobytes()))
        # PIL image mode: 1, L, P, I, F, RGB, YCbCr, RGBA, CMYK
        if pic.mode == "YCbCr":
            nchannel = 3
        elif pic.mode == "I;16":
            nchannel = 1
        else:
            nchannel = len(pic.mode)
        img = img.view(pic.size[1], pic.size[0], nchannel)
        # put it from HWC to CHW format
        # yikes, this transpose takes 80% of the loading time/CPU
        img = img.transpose(0, 1).transpose(0, 2).contiguous()
        if isinstance(img, torch.ByteTensor):
            return img.float().div(self.norm_value)
        else:
            return img

    def randomize_parameters(self):
        pass



from .model import DPTLarge
import torch
from jaseci.jsorc.live_actions import jaseci_action
import traceback
from fastapi import HTTPException
from PIL import Image
import base64
import io


@jaseci_action(act_group=["dpt"], allow_remote=True)
def setup(model: str = "dpt-large", device: str = None):
    global detector
    if device is None:
        _device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    else:
        _device = torch.device(device)
    detector = DPTLarge(device=_device, model=model)


@jaseci_action(act_group=["dpt"], allow_remote=True)
def estimate(image: str, b64: bool = False) -> str:
    try:
        _image = (
            Image.open(io.BytesIO(base64.b64decode(image)))
            if b64
            else Image.open(image)
        )
        depth_img = detector.estimate(_image).convert(mode="1").tobitmap()
        return base64.b64encode(depth_img).decode("utf-8")
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["dpt"], allow_remote=True)
def estimate_batch(images: list, b64: bool = False) -> list:
    try:
        _images = (
            [Image.open(io.BytesIO(base64.b64decode(image))) for image in images]
            if b64
            else [Image.open(image) for image in images]
        )
        depth_imgs = detector.estimate_batch(_images)
        return [
            base64.b64encode(depth_img.convert(mode="1").tobitmap()).decode("utf-8")
            for depth_img in depth_imgs
        ]
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.jsorc.remote_actions import launch_server

    launch_server(port=8000)



from .model import YolosDetector
import torch
from jaseci.jsorc.live_actions import jaseci_action
import traceback
from fastapi import HTTPException
import base64
import io
from PIL import Image


@jaseci_action(act_group=["yolos"], allow_remote=True)
def setup(device: str = None):
    global detector
    if device is None:
        _device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    else:
        _device = torch.device(device)
    detector = YolosDetector(device=_device)


@jaseci_action(act_group=["yolos"], allow_remote=True)
def detect(image, b64: bool = False, threshold: float = 0.5) -> list:
    try:
        if threshold > 1 or threshold < 0:
            raise ValueError("Threshold must be between 0 and 1")
        _image = (
            Image.open(io.BytesIO(base64.b64decode(image)))
            if b64
            else Image.open(image)
        )
        return detector.detect(_image, threshold)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["yolos"], allow_remote=True)
def detect_batch(images, b64: bool = False, threshold: float = 0.5) -> list:
    try:
        if threshold > 1 or threshold < 0:
            raise ValueError("Threshold must be between 0 and 1")
        _images = (
            [Image.open(io.BytesIO(base64.b64decode(image))) for image in images]
            if b64
            else [Image.open(image) for image in images]
        )
        return detector.detect_batch(_images, threshold)
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


@jaseci_action(act_group=["yolos"], allow_remote=True)
def get_labels() -> list:
    return detector.get_labels()


if __name__ == "__main__":
    from jaseci.jsorc.remote_actions import launch_server

    launch_server(port=8000)



from setuptools import setup, find_packages
from os.path import join

MODULES = ["rftm", "yolos", "dpt", "detr"]


def get_ver():
    with open(join("./jac_vision", "VERSION")) as version_file:
        return version_file.read().strip()


def get_extras_requires():
    extras_requires = {"all": []}
    for module in MODULES:
        with open(join("./jac_vision", module, "requirements.txt")) as req_file:
            extras_requires[module] = req_file.read().splitlines()
            extras_requires["all"].extend(extras_requires[module])
    return extras_requires


setup(
    name="jac_vision",
    version=get_ver(),
    packages=find_packages(include=["jac_vision", "jac_vision.*"]),
    install_requires=["jaseci", "pytest>=7.0.1,<7.1", "pytest-order>=1.0.1,<1.1"],
    extras_require=get_extras_requires(),
    package_data={
        "": ["*.json", "*.cfg", "VERSION", "*.yaml", "requirements.txt", "*.jac"],
    },
    author="Jason Mars",
    author_email="jason@jaseci.org",
    url="https://github.com/Jaseci-Labs/jaseci",
)



import torch
from PIL import Image
import torch
from jaseci.jsorc.live_actions import jaseci_action
import traceback
from fastapi import HTTPException
import base64
import io
import os
import requests
from .model import RFTM, ToTensor

RTFM_CLASSIFIER_WEIGHTS = "https://huggingface.co/spaces/jaseci/video-anomaly-detection/resolve/main/rftm_classifier.pth"
RTFM_RESNET_101_WEIGHTS = "https://huggingface.co/spaces/jaseci/video-anomaly-detection/resolve/main/rftm_resnet101.pth"


@jaseci_action(act_group=["rftm"], allow_remote=True)
def setup(device: str = None):
    global detector, _device
    os.makedirs("weights", exist_ok=True)
    if not os.path.exists(os.path.join("weights", "rftm_classifier.pth")):
        with open("weights/rftm_classifier.pth", "wb") as f:
            f.write(requests.get(RTFM_CLASSIFIER_WEIGHTS).content)
    if not os.path.exists(os.path.join("weights", "rftm_resnet101.pth")):
        with open("weights/rftm_resnet101.pth", "wb") as f:
            f.write(requests.get(RTFM_RESNET_101_WEIGHTS).content)

    _device = (
        torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if device is None
        else torch.device(device)
    )
    detector = RFTM(device=_device)


@jaseci_action(act_group=["rftm"], allow_remote=True)
def predict(frames: list, b64: bool = False) -> float:
    try:
        if b64:
            frames = [io.BytesIO(base64.b64decode(frame)) for frame in frames]
        if len(frames) != 16:
            raise ValueError("Must have 16 frames")
        inputs = torch.Tensor(1, 3, 16, 240, 320).to()
        for num, i in enumerate(frames):
            inputs[:, :, num, :, :] = ToTensor(1)(Image.open(i))
        with torch.no_grad():
            out = detector.predict(inputs)
        return out.item()
    except Exception as e:
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=str(e))


if __name__ == "__main__":
    from jaseci.jsorc.remote_actions import launch_server

    launch_server(port=8000)



from transformers import DetrImageProcessor, DetrForObjectDetection
import torch


class DetrDetector:
    def __init__(self, device=None, model="detr-resnet-50"):
        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device
        self.processor = DetrImageProcessor.from_pretrained(f"facebook/{model}")
        self.model = DetrForObjectDetection.from_pretrained(
            "facebook/detr-resnet-50"
        ).to(self.device)

    def detect(self, image, threshold=0.5):
        inputs = self.processor(images=image, return_tensors="pt").to(self.device)
        outputs = self.model(**inputs)

        target_sizes = torch.tensor([image.size[::-1]])
        results = self.processor.post_process_object_detection(
            outputs, threshold=threshold, target_sizes=target_sizes.to(self.device)
        )[0]

        detections = []
        for score, label, box in zip(
            results["scores"], results["labels"], results["boxes"]
        ):
            box = [round(i, 2) for i in box.tolist()]
            labelname = self.model.config.id2label[label.item()]
            score = round(score.item(), 3)
            detections.append({"label": labelname, "score": score, "box": box})

        return detections

    def detect_batch(self, images, threshold=0.5):
        inputs = self.processor(images=images, return_tensors="pt").to(self.device)
        outputs = self.model(**inputs)

        target_sizes = torch.tensor([image.size[::-1] for image in images])
        results = self.processor.post_process_object_detection(
            outputs, threshold=threshold, target_sizes=target_sizes.to(self.device)
        )

        batch_detections = []
        for i, image in enumerate(images):
            detections = []
            for score, label, box in zip(
                results[i]["scores"], results[i]["labels"], results[i]["boxes"]
            ):
                box = [round(i, 2) for i in box.tolist()]
                labelname = self.model.config.id2label[label.item()]
                score = round(score.item(), 3)
                detections.append({"label": labelname, "score": score, "box": box})
            batch_detections.append(detections)

        del inputs, outputs, results
        return batch_detections

    def get_labels(self):
        return self.model.config.id2label.values()



import numpy as np
import torch
from PIL import Image
from transformers import DPTFeatureExtractor, DPTForDepthEstimation


class DPTLarge:
    def __init__(self, device=None, model="dpt-large"):
        if device is None:
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = device
        self.feature_extractor = DPTFeatureExtractor.from_pretrained("Intel/dpt-large")
        self.model = DPTForDepthEstimation.from_pretrained("Intel/dpt-large").to(
            self.device
        )

    def estimate(self, image):
        inputs = self.feature_extractor(images=image, return_tensors="pt").to(
            self.device
        )
        outputs = self.model(**inputs)
        predicted_depth = outputs.predicted_depth

        prediction = torch.nn.functional.interpolate(
            predicted_depth.unsqueeze(1),
            size=image.size[::-1],
            mode="bicubic",
            align_corners=False,
        )

        output = prediction.squeeze().cpu().detach().numpy()
        formatted = (output * 255 / np.max(output)).astype("uint8")
        depth = Image.fromarray(formatted)

        return depth

    def estimate_batch(self, images):
        inputs = self.feature_extractor(images=images, return_tensors="pt").to(
            self.device
        )
        outputs = self.model(**inputs)
        predicted_depth = outputs.predicted_depth

        prediction = torch.nn.functional.interpolate(
            predicted_depth.unsqueeze(1),
            size=images[0].size[::-1],
            mode="bicubic",
            align_corners=False,
        )

        print(prediction.shape)
        output = prediction.squeeze().cpu().detach().numpy()
        formatted = (output * 255 / np.max(output)).astype("uint8")
        depth_imgs = [Image.fromarray(formatted[i]) for i in range(len(formatted))]

        del inputs, outputs
        return depth_imgs

    def get_labels(self):
        return self.model.config.id2label.values()
